from typing import Dict, List

from cognite.client.data_classes._base import CogniteFilter, CogniteResource, CogniteResourceList, CogniteUpdate


class Annotation(CogniteResource):
    """Representation of an annotation in CDF.

    Args:
        id (int, optional): [description]. Defaults to None.
        annotated_resource_id (int, optional): [description]. Defaults to None.
        text (str, optional): [description]. Defaults to None.
        annotated_resource_external_id (str, optional): [description]. Defaults to None.
        annotated_resource_type (str, optional): [description]. Defaults to None.
        linked_resource_id (int, optional): [description]. Defaults to None.
        linked_resource_external_id (str, optional): [description]. Defaults to None.
        linked_resource_type (str, optional): [description]. Defaults to None.
        annotation_type (str, optional): [description]. Defaults to None.
        status (str, optional): [description]. Defaults to None.
        source (str, optional): [description]. Defaults to None.
        region ([type], optional): [description]. Defaults to None.
        data ([type], optional): [description]. Defaults to None.
        created_time (int, optional): Time, in milliseconds since Jan. 1, 1970, when this annotation was created in CDF
        last_updated_time (int, optional): Time, in milliseconds since Jan. 1, 1970, when this annotation was last updated in CDF.
        cognite_client ([type], optional): The client to associate with this object.
    """

    def __init__(
        self,
        id: int = None,
        annotated_resource_id: int = None,
        text: str = None,
        annotated_resource_external_id: str = None,
        annotated_resource_type: str = None,
        linked_resource_id: int = None,
        linked_resource_external_id: str = None,
        linked_resource_type: str = None,
        annotation_type: str = None,
        status: str = None,
        source: str = None,
        region=None,
        data=None,
        created_time: int = None,
        last_updated_time: int = None,
        cognite_client=None,
    ):
        self.id = id
        self.annotated_resource_id = annotated_resource_id
        self.text = text
        self.annotated_resource_external_id = annotated_resource_external_id
        self.annotated_resource_type = annotated_resource_type
        self.linked_resource_id = linked_resource_id
        self.linked_resource_external_id = linked_resource_external_id
        self.linked_resource_type = linked_resource_type
        self.annotation_type = annotation_type
        self.status = status
        self.source = source
        self.region = region
        self.data = data
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self._cognite_client = cognite_client


class AnnotationFilter(CogniteFilter):
    def __init__(
        self,
        annotated_resource_type: str,
        annotated_resource_ids: List[Dict] = None,
        linked_resource_type: str = None,
        linked_resource_ids: List[Dict] = None,
        annotation_type: str = None,
        text: str = None,
        status: str = None,
        source: str = None,
    ):
        self.annotated_resource_type = annotated_resource_type
        self.annotated_resource_ids = annotated_resource_ids
        self.linked_resource_type = linked_resource_type
        self.linked_resource_ids = linked_resource_ids
        self.annotation_type = annotation_type
        self.text = text
        self.status = status
        self.source = source


class AnnotationUpdate(CogniteUpdate):
    pass


class AnnotationList(CogniteResourceList):
    _RESOURCE = Annotation
    _UPDATE = AnnotationUpdate



from cognite.client.data_classes._base import *


class IntegrationRun(CogniteResource):
    """A representation of a IntegrationRun.

    Args:
        external_id (str): The external ID of related integration provided by the client. Must be unique for the resource type.
        status (str): success/failure/seen.
        message (str): failure message.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        cognite_client (CogniteClient): The client to associate with this object.
    """

    def __init__(
        self,
        external_id: str = None,
        status: str = None,
        message: str = None,
        created_time: int = None,
        cognite_client=None,
    ):
        setattr(self, "external_id", external_id)
        setattr(self, "status", status)
        setattr(self, "message", message)
        setattr(self, "created_time", created_time)
        self._cognite_client = cognite_client


class IntegrationWithStatusesUpdate(CogniteUpdate):
    class _PrimitiveIntegrationWithStatusesUpdate(CognitePrimitiveUpdate):
        def set(self, value: Any) -> "IntegrationWithStatusesUpdate":
            return self._set(value)


class IntegrationRunUpdate(CogniteUpdate):
    class _PrimitiveIntegrationRunUpdate(CognitePrimitiveUpdate):
        def set(self, value: Any) -> "IntegrationRunUpdate":
            return self._set(value)


class IntegrationRunList(CogniteResourceList):
    _RESOURCE = IntegrationRun
    _UPDATE = IntegrationRunUpdate



import copy
from typing import Dict, List

import regex


def _label_groups(extractors: List[Dict], conditions: List[Dict]):
    colors = [
        "".join(f"{int(round(rgb*255)):02x}" for rgb in color)
        for color in [
            [0, 0.3470, 0.841],
            [0.9, 0.3250, 0.098],
            [0.9290, 0.694, 0.125],
            [0.4940, 0.184, 0.556],
            [0.4660, 0.674, 0.188],
            [0.3010, 0.745, 0.933],
            [0.6350, 0.078, 0.184],
        ]
    ]

    for extractor in extractors:
        extractor["groupLabel"] = {}

    for ci, condition in enumerate(conditions):
        if condition["conditionType"] != "equals":
            continue
        for ei, part in condition["arguments"]:
            extractor = extractors[ei]
            extractor["groupLabel"][part + 1] = ci  # 1-based

    for extractor in extractors:
        if extractor["extractorType"] != "regex":
            continue
        group_counter = 0

        def color_group(_):
            nonlocal group_counter, extractor
            group_counter += 1
            label_ix = extractor["groupLabel"].get(group_counter)
            if label_ix is not None:
                color = colors[label_ix % len(colors)]
                return f"<font color='#{color}'>\\g<{group_counter}></font>"
            else:
                return f"\\g<{group_counter}>"

        extractor["restorePattern"] = (
            "<font color='#666'>" + regex.sub(r"\(.*?\)", color_group, extractor["pattern"].strip("$^")) + "</font>"
        )

    return extractors


def _color_matches(extractors: List[Dict], matches: List[Dict]):
    columns = sorted(list({(extractor["entitySet"][:-1], extractor["field"]) for extractor in extractors}))  # order?
    patterns = {
        (extractor["entitySet"][:-1], extractor["field"]): extractor["pattern"].strip("^$")
        for extractor in extractors
        if extractor["extractorType"] == "regex"
    }
    html = ""
    html += (
        "<table><tr>"
        + "".join(
            f"<th style='text-align: center'><span style='font-size:150%'>{source_target}:{field}</span>"
            + f"<br>{patterns.get((source_target,field),'')}</th>"
            for source_target, field in columns
        )
        + "<th style='text-align: center'>Existing<br>Match Type</th>"
        + "<th style='text-align: center'>Consistent<br>Match?</th></tr>"
    )

    for match in matches:
        formatted = {
            "source": copy.copy(match.get("source")),
            "target": copy.copy(match.get("target")),
        }

        for extractor in extractors:
            if extractor["extractorType"] != "regex":
                continue
            source_target = extractor["entitySet"][:-1]  # singular
            field = extractor["field"]
            regex_match = regex.match(extractor["pattern"], match.get(source_target, {}).get(field, ""))
            if not regex_match:
                print(
                    "Unexpected lack of match of ", extractor["pattern"], match.get(source_target), field,
                )
                continue
            formatted_field = regex_match.expand(extractor["restorePattern"])
            formatted[source_target][field] = formatted_field

        html += (
            "<tr style='text-align: left'>"
            + "".join(
                f"<td style='text-align: left'>{formatted[source_target][field] }</td>"
                for source_target, field in columns
            )
            + f"<td>{match.get('existingMatchType')}</td><td>{match.get('consistentMatch')}</td></tr>"
        )
    return html



from typing import Dict, List, Tuple, Union

import requests
from cognite.client import utils
from cognite.client._api_client import APIClient

from cognite.experimental.data_classes.unstructured import UnstructuredAggregateList, UnstructuredSearchResultList


class GrepAPI(APIClient):
    _GREP_RESOURCE_PATH = "/files/unstructured"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._DPS_LIMIT = 10000

    def search(
        self, query: str, highlight: bool = False, filter: Dict = None, aggregates: Dict = None, limit: int = None,
    ) -> UnstructuredSearchResultList:
        """Grep stuff"""
        if limit == -1 or limit == float("inf"):
            limit = None
        body = {
            "filter": filter,
            "search": {"query": query, "highlight": highlight},
            "aggregates": aggregates,
            "limit": limit,
        }
        resp = self._post(url_path=self._GREP_RESOURCE_PATH + "/search", json=body)
        data = resp.json()
        items = data["items"]
        search_results = UnstructuredSearchResultList._load(items, cognite_client=self._cognite_client)
        if aggregates:
            search_results.aggregates = UnstructuredAggregateList._load(
                data["aggregates"], cognite_client=self._cognite_client
            )

        return search_results

    def download(self, id: int = None, external_id: str = None) -> str:
        """Download OCR results"""

        utils._auxiliary.assert_exactly_one_of_id_or_external_id(id, external_id)
        resp = self._post(
            url_path=self._GREP_RESOURCE_PATH + "/downloadlink/parsed",
            json={"items": [{"id": id, "externalId": external_id}]},
        )
        download_url = resp.json()["items"][0]["downloadUrl"]
        return requests.get(download_url).text



from typing import *

from cognite.client import utils
from cognite.client._api_client import APIClient

from cognite.experimental.data_classes import IntegrationRun, IntegrationRunList


class IntegrationsRunsAPI(APIClient):
    _RESOURCE_PATH = "/integrations/runs"
    _LIST_CLASS = IntegrationRunList

    def list(self, external_id: str, limit: int = 25) -> IntegrationRunList:
        """`List of runs for integration with given external_id <>`_

        Args:
            external_id (str): Integration external Id.
            limit (int, optional): Maximum number of integrations to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            IntegrationRunList: List of requested integration runs

        Examples:

            List integrationRuns::

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> runsList = c.integration_runs.list(external_id="test ext id", limit=5)
        """

        return self._list(method="GET", limit=limit, filter={"externalId": external_id})

    def create(self, run: Union[IntegrationRun, List[IntegrationRun]]) -> Union[IntegrationRun, IntegrationRunList]:
        """`Create one or more integrationRuns. <>`_

        You can create an arbitrary number of integrationRuns, and the SDK will split the request into multiple requests.

        Args:
            run (Union[IntegrationRun, List[IntegrationRun]]): IntegrationRun or list of integrationRuns to create.

        Returns:
            Union[IntegrationRun, IntegrationRunList]: Created integrationRun(s)

        Examples:

            Create new integrationRuns::

                >>> from cognite.experimental import CogniteClient
                >>> from cognite.experimental.data_classes import IntegrationRun
                >>> c = CogniteClient()
                >>> integrationRuns = [IntegrationRun(status="success", external_id="extId"),...]
                >>> res = c.integration_runs.create(integrationRuns)
        """
        utils._auxiliary.assert_type(run, "run", [IntegrationRun, list])
        return self._create_multiple(run)



from typing import Dict, List, Union

from cognite.client import utils
from cognite.client._api_client import APIClient
from cognite.client.data_classes import ContextualizationJob

from cognite.experimental._context_client import ContextAPI
from cognite.experimental.data_classes.templates import *


class TemplatesAPI(APIClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.groups = TemplateGroupsAPI(*args, **kwargs)
        self.versions = TemplateGroupVersionsAPI(*args, **kwargs)
        self.instances = TemplateInstancesAPI(*args, **kwargs)
        ctx_kwargs = {**kwargs, "api_version": "playground"}
        self.completion = TemplateCompletionAPI(*args, **ctx_kwargs)

    def graphql_query(self, external_id: str, version: int, query: str) -> GraphQlResponse:
        """
        `Run a GraphQL Query.`
        To learn more, see https://graphql.org/learn/

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group to run the query on.
            query (str): The GraphQL query to run.

        Returns:
            GraphQlResponse: the result of the query.

        Examples:
            Run a GraphQL query:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> query = '''
                >>>    {
                >>>        countryList {
                >>>           name,
                >>>           demographics {
                >>>               populationSize,
                >>>               growthRate
                >>>           },
                >>>           deaths {
                >>>               datapoints(limit: 100) {
                >>>                   timestamp,
                >>>                   value
                >>>               }
                >>>           }
                >>>        }
                >>>    }
                >>>    '''
                >>> result = c.templates.query.graphql_query("template-group-ext-id", 1, query)
        """
        path = "/templategroups/{}/versions/{}/graphql"
        path = utils._auxiliary.interpolate_and_url_encode(path, external_id, version)
        response = self._post(path, {"query": query})
        return GraphQlResponse._load(response.json())


class TemplateCompletionAPI(ContextAPI):

    _RESOURCE_PATH = "/context/schemas"

    def complete_type(self, external_id: str) -> ContextualizationJob:
        """(DEPRECATED) Completes a schema uploaded in CDF as a type.

        Args:
            external_id (str): External ID of the type to be completed

        Returns:
            ContextualizationJob: Resulting queued job. Note that .results property of this job will block waiting for results."""
        return self._run_job(job_path="/type", status_path="/", external_id=external_id)

    def complete(
        self, external_id: str, template_name: str, asset_property: str = None, version: int = None
    ) -> ContextualizationJob:
        """Completes a schema uploaded in CDF as a domain.

        Args:
            external_id (str): External ID of the template group to work on.
            template_name (str): Name of the template to be completed within the template group
            asset_property (str): Which field (with constant type) in the template defines the externalId of the parent asset in each entry. If ommitted, it is assumed the externalId of the template instances is the same as the parent asset's externalId.
            version (int): Version of the domain, can be ommitted to use the last one.

        Returns:
            ContextualizationJob: Resulting queued job. Note that .results property of this job will block waiting for results.

        Examples:
            Get template groups by external id:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.templates.completion.complete(external_id="abc",template_name="covid")
        """
        return self._run_job(
            job_path="/template",
            status_path="/",
            external_id=external_id,
            template_name=template_name,
            asset_property=asset_property,
            version=version,
        )


class TemplateGroupsAPI(APIClient):
    _RESOURCE_PATH = "/templategroups"
    _LIST_CLASS = TemplateGroupList

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create(
        self, template_groups: Union[TemplateGroup, List[TemplateGroup]]
    ) -> Union[TemplateGroup, TemplateGroupList]:
        """`Create one or more template groups.`

        Args:
            template_groups (Union[TemplateGroup, List[TemplateGroup]])

        Returns:
            Union[TemplateGroup, TemplateGroupList]: Created template group(s)

        Examples:
            create a new template group:

                >>> from cognite.experimental import CogniteClient
                >>> from cognite.experimental.data_classes import TemplateGroup
                >>> c = CogniteClient()
                >>> template_group_1 = TemplateGroup("sdk-test-group", "This is a test group")
                >>> template_group_2 = TemplateGroup("sdk-test-group-2", "This is another test group")
                >>> c.templates.groups.create([template_group_1, template_group_2])
        """
        return self._create_multiple(items=template_groups)

    def upsert(
        self, template_groups: Union[TemplateGroup, List[TemplateGroup]]
    ) -> Union[TemplateGroup, TemplateGroupList]:
        """`Upsert one or more template groups.`
        Will overwrite existing template group(s) with the same external id(s).

        Args:
            template_groups (Union[TemplateGroup, List[TemplateGroup]])

        Returns:
            Union[TemplateGroup, TemplateGroupList]: Created template group(s)

        Examples:
            create a new template group:

                >>> from cognite.experimental import CogniteClient
                >>> from cognite.experimental.data_classes import TemplateGroup
                >>> c = CogniteClient()
                >>> template_group_1 = TemplateGroup("sdk-test-group", "This is a test group")
                >>> template_group_2 = TemplateGroup("sdk-test-group-2", "This is another test group")
                >>> c.templates.groups.upsert([template_group_1, template_group_2])
        """
        path = self._RESOURCE_PATH + "/upsert"
        is_single = not isinstance(template_groups, list)
        if is_single:
            template_groups = [template_groups]
        updated = self._post(path, {"items": [item.dump(camel_case=True) for item in template_groups]},).json()["items"]
        res = self._LIST_CLASS._load(updated, cognite_client=self._cognite_client)
        if is_single:
            return res[0]
        return res

    def retrieve_multiple(self, external_ids: List[str], ignore_unknown_ids: bool = False) -> TemplateGroupList:
        """`Retrieve multiple template groups by external id.`

        Args:
            external_ids (List[str]): External IDs
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            TemplateGroupList: The requested template groups.

        Examples:
            Get template groups by external id:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.templates.groups.retrieve_multiple(external_ids=["abc", "def"])
        """
        return self._retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=ignore_unknown_ids, wrap_ids=True)

    def list(self, limit: int = 25, owners: List[str] = None) -> TemplateGroupList:
        """`Lists template groups stored in the project based on a query filter given in the payload of this request.`
        Up to 1000 template groups can be retrieved in one operation.

        Args:
            owners (List[str]): Include template groups that have any of these values in their `owner` field.
            limit (int): Maximum number of template groups to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            TemplateGroupList: List of requested template groups

        Examples:
            List template groups:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> template_group_list = c.templates.groups.list(limit=5)
        """
        filter = {}
        if owners is not None:
            filter["owners"] = owners
        return self._list(method="POST", limit=limit, filter=filter, partitions=None, sort=None)

    def delete(
        self, external_ids: Union[str, List[str]], ignore_unknown_ids: bool = False
    ) -> Union[TemplateGroup, TemplateGroupList]:
        """`Delete one or more template groups.`

        Args:
            external_ids (Union[str, List[str]]): External ID or list of external ids
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            None

        Examples:
            Delete template groups by external id:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.groups.delete(external_id=["a", "b"])
        """
        return self._delete_multiple(
            True, external_ids=external_ids, extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids}
        )


class TemplateGroupVersionsAPI(APIClient):
    _RESOURCE_PATH = "/templategroups/{}/versions"
    _LIST_CLASS = TemplateGroupVersionList

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def upsert(self, external_id: str, version: TemplateGroupVersion):
        """`Upsert a template group version.`
        A Template Group update supports specifying different conflict modes, which is used when an existing schema already exists.

        Patch -> It diffs the new schema with the old schema and fails if there are breaking changes.
        Update -> It sets the new schema as schema of a new version.
        Force -> It ignores breaking changes and replaces the old schema with the new schema.
        The default mode is "patch".

        Args:
            external_id (str): The external id of the template group.
            version (TemplateGroupVersion): The GraphQL schema of this version.

        Returns:
            TemplateGroupVersion: Created template group version

        Examples:
            create a new template group version modeling Covid-19:

                >>> from cognite.experimental import CogniteClient
                >>> from cognite.experimental.data_classes import TemplateGroup
                >>> c = CogniteClient()
                >>> template_group = TemplateGroup("sdk-test-group", "This template group models Covid-19 spread")
                >>> c.templates.groups.create(template_group)
                >>> schema = '''
                >>>     type Demographics @template {
                >>>         "The amount of people"
                >>>         populationSize: Int,
                >>>         "The population growth rate"
                >>>         growthRate: Float,
                >>>     }
                >>>     type Country @template {
                >>>         name: String,
                >>>         demographics: Demographics,
                >>>         deaths: TimeSeries,
                >>>         confirmed: TimeSeries,
                >>>     }'''
                >>> template_group_version = TemplateGroupVersion(schema)
                >>> c.templates.versions.upsert(template_group.external_id, template_group_version)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id) + "/upsert"
        version = self._post(resource_path, version.dump(camel_case=True),).json()
        return TemplateGroupVersion._load(version)

    def list(
        self, external_id: str, limit: int = 25, min_version: Optional[int] = None, max_version: Optional[int] = None
    ) -> TemplateGroupVersionList:
        """`Lists versions of a specified template group.`
        Up to 1000 template group version can be retrieved in one operation.

        Args:
            external_id (str): The external id of the template group.
            limit (int): Maximum number of template group versions to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            min_version: (Optional[int]): Exclude versions with a version number smaller than this.
            max_version: (Optional[int]): Exclude versions with a version number larger than this.

        Returns:
            TemplateGroupVersionList: List of requested template group versions

        Examples:
            List template group versions:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> template_group_list = c.templates.versions.list("template-group-ext-id", limit=5)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id)
        filter = {}
        if min_version is not None:
            filter["minVersion"] = min_version
        if max_version is not None:
            filter["maxVersion"] = max_version
        return self._list(resource_path=resource_path, method="POST", limit=limit, filter=filter)

    def delete(self, external_id: str, version: int) -> None:
        """`Delete a template group version.`

        Args:
            external_id (Union[str, List[str]]): External ID of the template group.
            version (int): The version of the template group to delete.

        Returns:
            None

        Examples:
            Delete template groups by external id:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.versions.delete("sdk-test-group", 1)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id)
        self._post(resource_path + "/delete", {"version": version})


class TemplateInstancesAPI(APIClient):
    _RESOURCE_PATH = "/templategroups/{}/versions/{}/instances"
    _LIST_CLASS = TemplateInstanceList

    def create(
        self, external_id: str, version: int, instances: Union[TemplateInstance, List[TemplateInstance]]
    ) -> Union[TemplateInstance, TemplateInstanceList]:
        """`Create one or more template instances.`

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group to create instances for.
            instances (Union[TemplateInstance, List[TemplateInstance]]): The instances to create.

        Returns:
            Union[TemplateInstance, TemplateInstanceList]: Created template instance(s).

        Examples:
            create new template instances for Covid-19 spread:

                >>> from cognite.experimental import CogniteClient
                >>> from cognite.experimental.data_classes import TemplateInstance
                >>> c = CogniteClient()
                >>> template_instance_1 = TemplateInstance(
                >>>                               external_id="norway",
                >>>                               template_name="Country",
                >>>                               field_resolvers={
                >>>                                   "name": ConstantResolver("Norway"),
                >>>                                   "demographics": ConstantResolver("norway_demographics"),
                >>>                                   "deaths": ConstantResolver("Norway_deaths"),
                >>>                                   "confirmed": ConstantResolver("Norway_confirmed"),
                >>>                                   }
                >>>                               )
                >>> template_instance_2 = TemplateInstance(
                >>>                               external_id="norway_demographics",
                >>>                               template_name="Demographics",
                >>>                               field_resolvers={
                >>>                                   "populationSize": ConstantResolver(5328000),
                >>>                                   "growthRate": ConstantResolver(value=0.02)
                >>>                                   }
                >>>                               )
                >>> c.templates.instances.create("sdk-test-group", 1, [template_instance_1, template_instance_2])
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._create_multiple(resource_path=resource_path, items=instances)

    def upsert(
        self, external_id: str, version: int, instances: Union[TemplateGroup, List[TemplateGroup]]
    ) -> Union[TemplateInstance, TemplateInstanceList]:
        """`Upsert one or more template instances.`
        Will overwrite existing instances.

        Args:
         external_id (str): The external id of the template group.
         version (int): The version of the template group to create instances for.
         instances (Union[TemplateInstance, List[TemplateInstance]]): The instances to create.

        Returns:
         Union[TemplateInstance, TemplateInstanceList]: Created template instance(s).

        Examples:
         create new template instances for Covid-19 spread:

             >>> from cognite.experimental import CogniteClient
             >>> from cognite.experimental.data_classes import TemplateInstance
             >>> c = CogniteClient()
             >>> template_instance_1 = TemplateInstance(
             >>>        external_id="norway",
             >>>        template_name="Country",
             >>>        field_resolvers={
             >>>            "name": ConstantResolver("Norway"),
             >>>            "demographics": ConstantResolver("norway_demographics"),
             >>>            "deaths": ConstantResolver("Norway_deaths"),
             >>>            "confirmed": ConstantResolver("Norway_confirmed"),
             >>>        }
             >>>    )
             >>> template_instance_2 = TemplateInstance(external_id="norway_demographics",
             >>>       template_name="Demographics",
             >>>       field_resolvers={
             >>>           "populationSize": ConstantResolver(5328000),
             >>>           "growthRate": ConstantResolver(0.02)
             >>>           }
             >>>   )
             >>> c.templates.instances.upsert("sdk-test-group", 1, [template_instance_1, template_instance_2])
        """
        if isinstance(instances, TemplateInstance):
            instances = [instances]
        resource_path = (
            utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version) + "/upsert"
        )
        updated = self._post(
            resource_path, {"items": [instance.dump(camel_case=True) for instance in instances]},
        ).json()["items"]
        res = TemplateInstanceList._load(updated, cognite_client=self._cognite_client)
        if len(res) == 1:
            return res[0]
        return res

    def retrieve_multiple(
        self, external_id: str, version: int, external_ids: List[str], ignore_unknown_ids: bool = False
    ) -> TemplateInstanceList:
        """`Retrieve multiple template instances by external id.`

        Args:
            external_id (str): The template group to retrieve instances from.
            version (int): The version of the template group.
            external_ids (List[str]): External IDs of the instances.
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            TemplateInstanceList: The requested template groups.

        Examples:
            Get template instances by external id:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.templates.instances.retrieve_multiple(external_id="sdk-test-group", version=1, external_ids=["abc", "def"])
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._retrieve_multiple(
            resource_path=resource_path, external_ids=external_ids, ignore_unknown_ids=ignore_unknown_ids, wrap_ids=True
        )

    def list(
        self,
        external_id: str,
        version: int,
        limit: int = 25,
        data_set_ids: Optional[List[int]] = None,
        template_names: Optional[List[str]] = None,
    ) -> TemplateInstanceList:
        """`Lists instances in a template group.`
        Up to 1000 template instances can be retrieved in one operation.

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group.
            limit (int): Maximum number of template group versions to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            data_set_ids: (Optional[List[int]]): Only include instances which has one of these values in their `data_set_id` field.
            template_names: (Optional[List[str]]): Only include instances which has one of these values in their `template_name` field.

        Returns:
            TemplateInstanceList: List of requested template instances

        Examples:
            List template instances:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> template_instances_list = c.templates.instances.list("template-group-ext-id", 1, limit=5)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        filter = {}
        if data_set_ids is not None:
            filter["dataSetIds"] = data_set_ids
        if template_names is not None:
            filter["template_names"] = template_names
        return self._list(resource_path=resource_path, method="POST", limit=limit, filter=filter)

    def delete(self, external_id: str, version: int, external_ids: List[str], ignore_unknown_ids: bool = False) -> None:
        """`Delete one or more template instances.`

        Args:
            external_id (Union[str, List[str]]): External ID of the template group.
            version (int): The version of the template group.
            external_ids (List[str]): The external ids of the template instances to delete
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            None

        Examples:
            Delete template groups by external id:

                >>> from cognite.experimental import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.instances.delete("sdk-test-group", 1, external_id=["a", "b"])
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._delete_multiple(
            resource_path=resource_path,
            external_ids=external_ids,
            wrap_ids=True,
            extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids},
        )



from typing import Dict, List, Union

from cognite.client import utils

from cognite.experimental._context_client import ContextAPI
from cognite.experimental.data_classes import EntityMatchingMatchRuleList, MatchRulesApplyJob, MatchRulesSuggestJob


class MatchRulesAPI(ContextAPI):
    _RESOURCE_PATH = "/context/matchrules"

    def apply(
        self, sources: List[dict], targets: List[dict], rules: Union[List[dict], EntityMatchingMatchRuleList]
    ) -> MatchRulesApplyJob:
        """Apply match rules with priorities to match source entities with target entities.

        Args:
            sources (List[dict]): List of source entities in json format.
            targets (List[dict]): List of target entities in json format.
            rules (Union[List[dict], EntityMatchingMatchruleList]): List of match rules with priorities to apply to the
                entities
        Returns:
            MatchRulesApplyJob: Job, calling .rules waits for completion."""

        if isinstance(rules, EntityMatchingMatchRuleList):
            rules = [
                {k: r.dump()[k] for k in ["extractors", "conditions", "priority"] if r.dump().get(k)}
                for r in rules.data
            ]

        return self._run_job(
            job_path="/apply",
            status_path="/apply/",
            sources=sources,
            targets=targets,
            rules=rules,
            job_cls=MatchRulesApplyJob,
        )

    def suggest(self, sources: List[dict], targets: List[dict], matches: List[dict]) -> MatchRulesSuggestJob:
        """Suggest match rules with priorities based on existing matches between source and target entities.

        Args:
            sources (List[dict]): List of dict representation of source entities to suggest rules for.
            targets (List[dict]): List of dict representation of target entities to suggest rules for.
            matches (list[dict]): List of matches in terms of source_id or source_external_id and similar for target.

        Returns:
            MatchRulesSuggestJob: Resulting queued job. Note that .rules property of this job will block waiting for
            results.
        """
        matches = [{utils._auxiliary.to_camel_case(k): v for k, v in match.items()} for match in matches]
        return self._run_job(
            job_path="/suggest",
            status_path="/suggest/",
            sources=sources,
            targets=targets,
            matches=matches,
            job_cls=MatchRulesSuggestJob,
        )
